@@ -3,4 +3,4 @@
 from bibframe.plugin import linkreport
 from bibframe import register_service
 
-register_service(linkreport.linkreport)
+register_service(linkreport.PLUGIN_INFO)
